use std::io::Read;
use std::str::FromStr;
use std::fs::File;

use color_eyre::{eyre};
use eyre::eyre;


pub fn load(filename: &str) -> eyre::Result<String> {
    let mut file = File::open(filename)?;
    let mut data = String::new();
    file.read_to_string(&mut data)?;
    Ok(data)
}


pub struct LineParser<'a, T: FromStr> {
    it: std::str::Lines<'a>,
    convert: fn(&str) -> Result<T, <T as FromStr>::Err>
}


impl<'a, T: FromStr> Iterator for LineParser<'a, T> {
    type Item = Result<T, <T as FromStr>::Err>;

    fn next(&mut self) -> Option<Self::Item> {
        self.it.next().map(|s|
            (self.convert)(s)
        )
    }
}


// define a trait so that we can attach it to String
pub trait ParseLines {
    fn parse_lines<T: FromStr>(&self) -> LineParser<'_, T>;
}


impl ParseLines for str {
    fn parse_lines<T: FromStr>(&self) -> LineParser<'_, T> {
        LineParser {
            it: self.lines(),
            convert: str::parse::<T>
        }
    }
}


impl ParseLines for String {
    fn parse_lines<T: FromStr>(&self) -> LineParser<'_, T> {
        self[..].parse_lines::<T>()
    }
}


pub trait IterExt: Iterator {
    fn take_pair(&mut self) -> eyre::Result<(Self::Item, Self::Item)> {
        let a = match self.next() {
            Some(v) => v,
            None => return Err(eyre!("Not enough values to unpack")),
        };
        let b = match self.next() {
            Some(v) => v,
            None => return Err(eyre!("Not enough values to unpack")),
        };
        Ok((a, b))
    }
}

impl<I: Iterator> IterExt for I {}
